from functools import partial
from pathlib import Path
from typing import List, Optional

from loguru import logger
from randonneur_data import Registry
from tqdm import tqdm

from .edge_functions import (
    migrate_edges_create,
    migrate_edges_delete,
    migrate_edges_disaggregate,
    migrate_edges_replace,
    migrate_edges_update,
)
from .errors import WrongGraphContext
from .utils import FlexibleLookupDict, apply_mapping
from .config import MigrationConfig

verb_dispatch = {
    "create": migrate_edges_create,
    "replace": migrate_edges_replace,
    "delete": migrate_edges_delete,
    "update": migrate_edges_update,
    "disaggregate": migrate_edges_disaggregate,
}


def migrate_edges(
    graph: List[dict],
    migrations: dict,
    config: Optional[MigrationConfig] = None,
) -> List[dict]:
    """For each edge in each node in ``data``, check each transformation in ``migrations``. For each
    transformation for which there is a match, make the given changes to the edge.

    Here is an example:

    ```python
    migrate_edges(
        graph=[{"edges": [{"name": "foo"}]}],
        migrations={"update": [{"source": {"name": "foo"}, "target": {"location": "bar"}}]},
    )
    >>> [{"edges": [{"name": "foo", "location": "bar"}]}]
    ```

    The changes can be customized with a `MigrationConfig` object. See the `MigrationConfig` docs
    for information on its input arguments.

    Returns `graph` with altered content.

    """
    config = config or MigrationConfig()
    progressbar = (
        partial(tqdm, desc="Transforming graph nodes") if config.verbose else lambda x: iter(x)
    )

    if config.mapping:
        migrations = apply_mapping(
            migrations=migrations, mapping=config.mapping, verbs=config.verbs
        )

    verbs = list(filter(lambda x: x in verb_dispatch and x in migrations, config.verbs))
    logger.info("Can apply the following transformation verbs: {v}", v=verbs)

    flds = {
        verb: FlexibleLookupDict(
            input_data=migrations[verb],
            fields_filter=config.fields,
            case_sensitive=config.case_sensitive,
        )
        for verb in verbs
    }

    if "create" in migrations and "create" in config.verbs:
        logger.warning(
            """`migrations` has `create` section - this will add exchanges to all nodes.
This is almost never the desired behaviour, consider removing `create` from the `verb` input."""
        )

    for node in progressbar(graph):
        if config.node_filter and not config.node_filter(node):
            continue

        for verb in verbs:
            verb_dispatch[verb](
                node=node,
                migration_fld=flds[verb],
                edge_filter=config.edge_filter,
                edges_label=config.edges_label,
                verbose=config.verbose,
            )

    return graph


def migrate_edges_with_stored_data(
    graph: List[dict],
    label: str,
    data_registry_path: Optional[Path] = None,
    config: Optional[MigrationConfig] = None,
) -> List[dict]:
    """A simple wrapper to load from a `randonneur_data.Registry` with some basic sanity checks."""
    config = config or MigrationConfig()

    try:
        migrations = Registry(data_registry_path).get_file(label)
        logger.info(
            "Loaded transformation data {l} from registry with following verbs: {v}",
            l=label,
            v=[k for k in migrations if k in verb_dispatch],
        )
    except KeyError:
        raise KeyError(f"Transformation {label} not found in given transformation registry")

    if "edges" not in migrations.get("graph_context", []):
        raise WrongGraphContext(f"{label} migration can't be used on edges")

    return migrate_edges(
        graph=graph,
        migrations=migrations,
        config=config,
    )



from copy import deepcopy
from typing import Callable, List, Optional

from .utils import rescale_edge


def migrate_edges_create(
    node: dict,
    migration_fld: dict,
    edges_label: str,
    verbose: bool,
    edge_filter: Optional[Callable] = None,
) -> dict:
    pass


def migrate_edges_delete(
    node: dict,
    migration_fld: dict,
    edges_label: str,
    verbose: bool,
    edge_filter: Optional[Callable] = None,
) -> dict:
    pass


def migrate_edges_disaggregate(
    node: dict,
    migration_fld: dict,
    edges_label: str,
    verbose: bool,
    fields: Optional[List[str]] = None,
    edge_filter: Optional[Callable] = None,
    case_sensitive: bool = False,
) -> dict:
    edges_to_add, edges_to_remove = [], set()

    for edge in node.get(edges_label, []):
        if edge_filter and not edge_filter(edge):
            continue
        try:
            for allocated in migration_fld[edge]["targets"]:
                new_edge = rescale_edge(deepcopy(edge), allocated["allocation"])
                new_edge.update(allocated)
                edges_to_add.append(new_edge)
            edges_to_remove.add(id(edge))
        except KeyError:
            continue
    if edges_to_remove:
        node[edges_label] = [
            edge for edge in node[edges_label] if id(edge) not in edges_to_remove
        ] + edges_to_add

    return node


def migrate_edges_replace(
    node: dict,
    migration_fld: dict,
    edges_label: str,
    verbose: bool,
    fields: Optional[List[str]] = None,
    edge_filter: Optional[Callable] = None,
    case_sensitive: bool = False,
) -> dict:
    for edge in node.get(edges_label, []):
        if edge_filter and not edge_filter(edge):
            continue
        try:
            migration = migration_fld[edge]
            if "conversion_factor" in migration["target"]:
                rescale_edge(edge, migration["target"]["conversion_factor"])
            elif "allocation" in migration["target"]:
                rescale_edge(edge, migration["target"]["allocation"])
            edge.update(migration["target"])
        except KeyError:
            continue
    return node


def migrate_edges_update(
    node: dict,
    migration_fld: dict,
    edges_label: str,
    verbose: bool,
    fields: Optional[List[str]] = None,
    edge_filter: Optional[Callable] = None,
    case_sensitive: bool = False,
) -> dict:
    """Difference is in intent of data developer, not in implementation."""
    return migrate_edges_replace(
        node=node,
        migration_fld=migration_fld,
        fields=fields,
        edges_label=edges_label,
        verbose=verbose,
        edge_filter=edge_filter,
        case_sensitive=case_sensitive,
    )



import json
import os
from datetime import datetime, timezone
from pathlib import Path
from typing import Optional

from .validation import DatapackageMetadata, Contributors, MappingFields, validate_data_for_verb

CC_BY = [
    {
        "name": "CC BY 4.0",
        "path": "https://creativecommons.org/licenses/by/4.0/",
        "title": "Creative Commons Attribution 4.0 International",
    }
]


class Datapackage:
    def __init__(
        self,
        *,
        name: str,
        description: str,
        contributors: list,
        mapping_source: dict,
        mapping_target: dict,
        source_id: Optional[str] = None,
        target_id: Optional[str] = None,
        homepage: Optional[str] = None,
        created: Optional[datetime] = None,
        version: str = "1.0.0",
        licenses: Optional[list] = None,
        graph_context: Optional[list] = None,
    ):
        self.name = name
        self.description = description
        self.contributors = contributors
        self.source_id = source_id
        self.target_id = target_id
        self.homepage = homepage
        self.created = created or datetime.now(timezone.utc).isoformat()
        self.mapping = {"source": mapping_source, "target": mapping_target}
        self.licenses = licenses or CC_BY
        self.version = version
        self.data = {}
        self.graph_context = graph_context or ["edges"]

        for contributor in contributors:
            Contributors(**contributor)
        MappingFields(**mapping_source)
        MappingFields(**mapping_target)
        DatapackageMetadata(
            name=self.name,
            description=self.description,
            source_id=self.source_id,
            target_id=self.target_id,
            homepage=self.homepage,
            created=self.created,
            version=self.version,
            licenses=self.licenses,
            graph_context=self.graph_context,
        )

    def metadata(self) -> dict:
        data = {
            "name": self.name,
            "description": self.description,
            "contributors": self.contributors,
            "created": self.created.isoformat()
            if isinstance(self.created, datetime)
            else self.created,
            "version": self.version,
            "licenses": self.licenses,
            "graph_context": self.graph_context,
            "mapping": self.mapping,
            "source_id": self.source_id,
            "target_id": self.target_id,
        }
        if self.homepage:
            data["homepage"] = self.homepage
        return data

    def add_data(self, verb: str, data: list) -> None:
        validate_data_for_verb(verb=verb, data=data, mapping=self.mapping)
        if verb not in self.data:
            self.data[verb] = []
        self.data[verb].extend(data)

    def to_json(self, filepath: Path) -> Path:
        if not isinstance(filepath, Path):
            filepath = Path(filepath)
        if filepath.suffix.lower() != ".json":
            filepath = filepath.parent / f"{filepath.name}.json"
        if not os.access(filepath.parent, os.W_OK):
            raise OSError(f"Can't write to directory {filepath.parent}")

        with open(filepath, "w", encoding="utf-8") as f:
            json.dump(self.metadata() | self.data, f, indent=2, ensure_ascii=False)

        return filepath



from copy import deepcopy

from tqdm import tqdm

from .utils import matcher


def migrate_datasets(
    migration_data,
    lci_database,
    create=True,
    disaggregate=True,
    replace=True,
    update=True,
    delete=True,
    dataset_filter=None,
    verbose=False,
    only_one_change=True,
):
    """Migrate datasets from ``lci_database`` using the data in ``migration_data``.

    ``create``, ``disaggregate``, ``replace``, ``update``, and ``delete`` changes can be applied,
    and their activation is controlled by their respective flags. See the README for more detail
    on the specifics of these changes, and the data formats of ``migration_data`` and
    ``lci_database``.

    You can filter the datasets to be changed with ``dataset_filter``. If given, these should be
    callables that take a dataset as the single input argument, and return a ``True`` if changes
    should be made.

    ``verbose`` controls whether a progressbar is shown when iterating over ``lci_database``.

    ``only_one_change`` determines whether more than one change (either of multiple types, or of
    the same type if multiple changes which match the original exchange are given) is executed. Be
    very careful with this, many changes to large databases should be carefully checked.

    Returns ``lci_database`` with altered content.
    """
    progressbar = tqdm if verbose else lambda x: iter(x)

    datasets_to_delete = []
    datasets_to_add = migration_data.get("create-datasets", []) if create else []

    for dataset in progressbar(lci_database):
        found = False

        if dataset_filter is not None and not dataset_filter(dataset):
            continue

        for possible in migration_data.get("delete", []):
            if not (found and only_one_change) and matcher(possible, dataset):
                datasets_to_delete.append(dataset)
                found = True

        for possible in migration_data.get("replace", []):
            if not (found and only_one_change) and matcher(possible["source"], dataset):
                datasets_to_add.append(deepcopy(possible["target"]))
                datasets_to_delete.append(dataset)
                found = True

        for possible in migration_data.get("update", []):
            if not (found and only_one_change) and matcher(possible["source"], dataset):
                dataset.update(possible["target"])
                found = True

        for possibles in migration_data.get("disaggregate", []):
            if not (found and only_one_change) and matcher(possibles["source"], dataset):
                for new_ds in possibles["targets"]:
                    old_ds = deepcopy(dataset)
                    old_ds.update(new_ds)
                    datasets_to_add.append(old_ds)
                datasets_to_delete.append(dataset)
                found = True

    if datasets_to_delete:
        lci_database = [dataset for dataset in lci_database if dataset not in datasets_to_delete]
    if datasets_to_add:
        lci_database.extend(datasets_to_add)

    return lci_database



import math
from collections.abc import Iterable, Mapping
from numbers import Number
from typing import Any, List, Optional

import stats_arrays as sa

from .errors import MultipleTransformations, ConflictingConversionFactors

ALL_VERBS = ["create", "delete", "replace", "update", "disaggregate"]
SAFE_VERBS = ["update", "replace", "disaggregate"]


def apply_mapping(migrations: dict, mapping: dict, verbs: List[str]) -> dict:
    """Apply the label changes in `mapping` to the transformations in `migrations`."""
    if "source" in mapping:
        for verb in verbs:
            for transformation in migrations.get(verb, []):
                for key, value in mapping["source"].items():
                    if key in transformation["source"]:
                        transformation["source"][value] = transformation["source"].pop(key)
    if "target" in mapping:
        for verb in verbs:
            if verb == "disaggregate":
                for transformation_list in migrations.get(verb, []):
                    for transformation in transformation_list["targets"]:
                        for key, value in mapping["target"].items():
                            if key in transformation:
                                transformation[value] = transformation.pop(key)
            elif verb in ("create", "delete"):
                continue
            else:
                for transformation in migrations.get(verb, []):
                    for key, value in mapping["target"].items():
                        if key in transformation["target"]:
                            transformation["target"][value] = transformation["target"].pop(key)
    return migrations


def rescale_edge(edge: dict, factor: Number) -> dict:
    """Rescale edges, including formulas and uncertainty values, by a constant factor"""
    if not isinstance(factor, Number):
        raise ValueError(f"Can't rescale by non-number `factor` {factor}")
    if edge.get("formula"):
        edge["formula"] = "({}) * {}".format(edge["formula"], factor)

    # Special cases
    if "uncertainty type" not in edge:
        if "amount" in edge:
            edge["amount"] *= factor
        return edge
    elif factor == 0:
        edge["amount"], edge["uncertainty type"] = 0, sa.UndefinedUncertainty.id
        for attr in ("minimum", "maximum", "scale", "shape", "loc"):
            if attr in edge:
                del edge[attr]
        return edge

    # stats_arrays uncertainty
    if edge["uncertainty type"] in (sa.UndefinedUncertainty.id, sa.NoUncertainty.id):
        edge["amount"] = edge["loc"] = factor * edge["amount"]
    elif edge["uncertainty type"] == sa.NormalUncertainty.id:
        edge["amount"] = edge["loc"] = factor * edge["amount"]
        edge["scale"] *= factor
    elif edge["uncertainty type"] == sa.LognormalUncertainty.id:
        edge["amount"] = factor * edge["amount"]
        edge["loc"] = math.log(abs(edge["amount"]))
        if edge["amount"] < 0:
            edge["negative"] = True
        elif "negative" in edge:
            del edge["negative"]
    elif edge["uncertainty type"] == sa.TriangularUncertainty.id:
        edge["minimum"] *= factor
        edge["maximum"] *= factor
        if edge["minimum"] > edge["maximum"]:
            edge["minimum"], edge["maximum"] = edge["maximum"], edge["minimum"]
        if "amount" in edge:
            edge["amount"] = edge["loc"] = factor * edge["amount"]
        else:
            edge["amount"] = edge["loc"] = (edge["minimum"] + edge["maximum"]) / 2
    elif edge["uncertainty type"] == sa.UniformUncertainty.id:
        edge["minimum"] *= factor
        edge["maximum"] *= factor
        if edge["minimum"] > edge["maximum"]:
            edge["minimum"], edge["maximum"] = edge["maximum"], edge["minimum"]
        if "amount" in edge:
            edge["amount"] *= factor
        else:
            edge["amount"] = edge["loc"] = (edge["minimum"] + edge["maximum"]) / 2
    else:
        raise ValueError(f"Edge can't be automatically rescaled:\n\t{edge}")
    return edge


def right_case(value: Any, case_sensitive: bool) -> Any:
    """Convert strings to lower case, and lists to tuples."""
    if isinstance(value, str) and not case_sensitive:
        return value.lower()
    elif isinstance(value, (tuple, list)) and not case_sensitive:
        return tuple([v.lower() if isinstance(v, str) else v for v in value])
    elif isinstance(value, (tuple, list)) and not case_sensitive:
        return tuple(value)
    else:
        return value


class FlexibleLookupDict(Mapping):
    def __init__(
        self,
        input_data: Iterable[dict],
        fields_filter: Optional[List[str]] = None,
        case_sensitive: bool = False,
    ):
        """A dictionary that allow for more flexible matching of dictionaries against other dicts.

        `input_data` is a dictionary like `{"foo": {"first": True, "bar": 42}}`. We want to match
        this input against `{'first': True}` and get back `foo`. Here is an examples:

        ```python
        fld = FlexibleLookupDict(
            input_data=[
                {"source": {"foo": "a", "bar": "b"}},
                {"source": {"foo": "b"}},
            ]
        )
        fld[{"foo": "b"}] == {"source": {"foo": "b"}}
        >>> True
        ```

        For real data we would have input data with both `source` and `target` (or `targets` for
        disaggregation) keys. This class makes the **strong assumption** that `input_data` has
        `source` and `target`/`targets` keys.

        We need to match a dictionary against another dictionary, but the other dictionary doesn't
        have a fixed set of keys - they can vary across all the possibilities. We therefore allow
        matching based on each unique combination of keys present.

        If `fields_filter` is given, then only consider keys present in that list.

        ```python
        fld = FlexibleLookupDict(
            input_data=[
                {"source": {"foo": "a", "bar": "b"}},
                {"source": {"foo": "b"}},
            ],
            fields_filter=["foo"]
        )
        fld[{"foo": "b", "other": "whatever"}] == {"source": {"foo": "b"}}
        >>> True
        ```

        If `case_sensitive`, then do case-sensitive matching on values (not keys) when comparing
        strings. Here is an example of a *case-insensitve* match:

        ```python
        fld = FlexibleLookupDict(
            input_data=[
                {"source": {"foo": "a", "bar": "b"}},
                {"source": {"foo": "b"}},
            ],
            case_sensitive=False
        )
        fld[{"foo": "B"}] == {"source": {"foo": "b"}}
        >>> True
        ```

        """
        self._case_sensitive = case_sensitive
        self._field_combinations = set()
        self._dict = {}

        if fields_filter:
            fields_filter = set(fields_filter)

        for obj in input_data:
            fields = set(obj["source"]).difference({"allocation"})
            if fields_filter:
                fields = fields.intersection(fields_filter)
            self._field_combinations.add(tuple(sorted(fields)))
            key = tuple(
                [right_case(obj["source"][field], case_sensitive) for field in sorted(fields)]
            )
            try:
                # Short circuits to `KeyError` if not present
                # We don't bother examining if disaggregation is consistent, this shouldn't be
                # allowed full stop
                if "targets" in obj and self._dict[key]:
                    raise MultipleTransformations(f"""
Found multiple transformations including disaggregation for:
{obj['source']}
""")

                # `obj` is already present in the dictionary. This is OK, if the *functionally
                # equivalent* values are being added. We would like to reject this data
                # completely, but then we have to fix the input data ourselves, and... yuck.
                existing = {
                    key: right_case(value, case_sensitive)
                    for key, value in self._dict[key]["target"].items()
                }
                given = {
                    key: right_case(value, case_sensitive) for key, value in obj["target"].items()
                }
                if existing != given:
                    raise MultipleTransformations(
                        f"""
Found multiple transformations for following field inputs:
{obj['source']}

Targets:

{obj['target']}

{self._dict[key]['target']}
"""
                    )
                if "conversion_factor" in obj and "conversion_factor" not in self._dict[key]:
                    self._dict[key]["conversion_factor"] = obj["conversion_factor"]
                elif "conversion_factor" in obj and "conversion_factor" in self._dict[key]:
                    if not math.isclose(
                        obj["conversion_factor"],
                        self._dict[key]["conversion_factor"],
                        abs_tol=1e-2,
                        rel_tol=1e-2,
                    ):
                        raise ConflictingConversionFactors(f"""
Found at least two different conversion factors for the same transformation.
First: {obj["conversion_factor"]}
Second: {self._dict[key]["conversion_factor"]}
For conversion from: {obj["source"]}
To: {obj["target"]}
                        """)
            except KeyError:
                self._dict[key] = obj

    def __getitem__(self, obj: dict) -> dict:
        if not isinstance(obj, dict):
            raise ValueError

        for field_combination in self._field_combinations:
            try:
                return self._dict[
                    tuple(
                        [
                            right_case(obj.get(field), self._case_sensitive)
                            for field in field_combination
                        ]
                    )
                ]
            except KeyError:
                continue

        raise KeyError

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> Iterable:
        return iter(self._dict)



from typing import Optional, Dict, Callable, List

from pydantic import BaseModel

from .utils import SAFE_VERBS


class MigrationConfig(BaseModel):
    """
    A class that stores configuration options for applying migrations.

    The following can be specified:

    `mapping`: Change the labels in the `migrations` data to match your data schema. `mapping`
    needs to be a dict with `source` and/or `target` keys, each value should be a dictionary with
    `{old_label: new_label}` pairs.

    ```python
    migrate_edges(
        graph=[{"edges": [{"name": "foo"}]}],
        migrations={"update": [{"source": {"not-name": "foo"}, "target": {"location": "bar"}}]},
        config=MigrationConfig(mapping={"source": {"not-name": "name"}})
    )
    >>> [{"edges": [{"name": "foo", "location": "bar"}]}]
    ```

    `node_filter`: A callable which determines if edges in that node should be modified. Returns
    `True` if the node *should be* modified.

    ```python
    migrate_edges(
        graph=[{"edges": [{"name": "foo"}]}],
        migrations={"update": [{"source": {"name": "foo"}, "target": {"location": "bar"}}]},
        config=MigrationConfig(node_filter=lambda node: node.get("sport") == "🏄‍♀️")
    )
    >>> [{"edges": [{"name": "foo"}]}]
    ```

    `edge_filter`: A callable which determines if a specific edge should be modified. Returns
    `True` if the edge *should be* modified.

    `fields`: A list of strings used when checking if the given transformation matches the edge
    under consideration. In other words, only use the fields in `fields` when checking the `source`
    values in each transformation for a match. Each field in `fields` doesn't have to be in each
    transformation.

    ```python
    migrate_edges(
        graph=[{"edges": [{"name": "foo"}]}],
        migrations={"update": [
            {"source": {"name": "foo", "missing": "🔍"}, "target": {"location": "bar"}}
        ]},
        config=MigrationConfig(fields=["name"]),
    )
    >>> [{"edges": [{"name": "foo", "location": "bar"}]}]
    ```

    `verbose`: Display progress bars and more logging messages.

    `edges_label`: The label used for edges in the nodes of the `graph`. Defaults to `edges`.

    ```python
    migrate_edges(
        graph=[{"e": [{"name": "foo"}]}],
        migrations={"update": [{"source": {"name": "foo"}, "target": {"location": "bar"}}]},
        config=MigrationConfig(edges_label="e"),
    )
    >>> [{"edges": [{"name": "foo", "location": "bar"}]}]
    ```

    `verbs`: The list of transformation types from `migrations` to apply. Transformations are run
    in the order as given in `verbs`, and in some complicated cases you may want to keep the same
    verbs but change their order to get the desired output state. In general, such complicated
    transformations should be broken down to smaller discrete and independent transformations
    whenever possible, and logs checked carefully after their application.

    Only the verbs `create`, `disaggregate`, `replace`, `update`, and `delete` are used in this
    function, regardless of what is given in `verbs`, as we don't know how to handle custom verbs.
    We need to write custom functions for each verb as they have difference behaviour.

    `case_sensitive`: Flag indicating whether to do case sensitive matching of transformations to
    nodes or edges in the graph. Default is false, as practical experience has shown us that cases
    get commonly changed by software developers or users. Only applies to string values.

    """

    mapping: Optional[Dict[str, dict]] = None
    node_filter: Optional[Callable] = None
    edge_filter: Optional[Callable] = None
    fields: Optional[List[str]] = None
    verbose: bool = False
    edges_label: str = "edges"
    verbs: List[str] = SAFE_VERBS
    case_sensitive: bool = False



import pytest
from pydantic import ValidationError

from randonneur import MigrationConfig


def test_migration_config_valid():
    assert MigrationConfig()
    assert MigrationConfig(
        mapping={"source": {"foo": "bar"}, "target": {"1": "2"}},
        node_filter=lambda x: True,
        edge_filter=lambda x: False,
        fields=["name", "reference product"],
        verbose=True,
        edges_label="this",
        verbs=["replace", "delete"],
        case_sensitive=True,
    )


def test_migration_config_invalid():
    with pytest.raises(ValidationError):
        MigrationConfig(mapping=["foo", "bar"])



from datetime import datetime
from typing import List, Dict, Optional, Union
from pprint import pformat

from pydantic import BaseModel, Field

from .errors import UnmappedData

VERBS = {"create", "replace", "update", "delete", "disaggregate"}


class Contributors(BaseModel):
    title: str
    role: str
    path: str


class MappingFields(BaseModel):
    expression_language: str = Field(alias="expression language")
    labels: Dict[str, Union[str, list]]


class DatapackageMetadata(BaseModel):
    """Validate given Datapackage metadata.

    Pydantic gives us nice error messages for free."""

    name: str
    description: str
    source_id: Optional[str] = None
    target_id: Optional[str] = None
    homepage: Optional[str] = None
    created: Optional[datetime] = datetime.now()
    version: str
    licenses: List[Dict[str, str]]
    graph_context: List[str]


def validate_data_for_verb(verb: str, data: list, mapping: dict) -> None:
    if verb not in VERBS:
        raise ValueError(f"Transformation verb {verb} must be one of {VERBS}")

    all_source_keys = set(mapping["source"]["labels"])
    all_target_keys = set(mapping["target"]["labels"])

    for element in data:
        missing = set(element["source"]).difference(all_source_keys)
        if missing:
            raise UnmappedData(
                f"""
One of more `source` data attributes is not found in given mapping:
Mapping has the following fields:
{all_source_keys}
However, the given data also includes:
{missing}
In the data object:
{pformat(element)}
"""
            )
        if verb == "disaggregate":
            missing = (
                set.union(*[set(obj) for obj in element["targets"]])
                .difference(all_target_keys)
                .difference({"allocation"})
            )
        elif verb == "delete":
            # No target element
            missing = None
        else:
            missing = (
                set(element["target"]).difference(all_target_keys).difference({"allocation"})
            )
        if missing:
            raise UnmappedData(
                f"""
One of more `target` data attributes is not found in given mapping:
Mapping has the following fields:
{all_target_keys}
However, the given data also includes:
{missing}
In the data object:
{pformat(element)}
"""
            )



class MappingConstants:
    SIMAPRO_CSV = {
        "expression language": "like JSONPath",
        "labels": {
            "identifier": 'Process[*]."Process identifier".text',
            "name": "Process[*].Products[*].text[0]",
            "platform_id": 'Process[*]."Platform Identifier"',
            "unit": [
                '["Emissions to air/", Process[*]."Emissions to air".[2]]',
                '["Emissions to soil/", Process[*]."Emissions to soil".[2]]',
                '["Emissions to water/", Process[*]."Emissions to water".[2]]',
                '["Resources/", Process[*]."Resources".[2]]',
            ],
            "context": [
                '["Emissions to air/", Process[*]."Emissions to air".[1]]',
                '["Emissions to soil/", Process[*]."Emissions to soil".[1]]',
                '["Emissions to water/", Process[*]."Emissions to water".[1]]',
                '["Resources/", Process[*]."Resources".[1]]',
            ],
        },
    }
    ECOSPOLD2 = {
        "expression language": "XPath",
        "labels": {
            "filename": "concat(//*:activity/@id, '_', //*:intermediateExchange[*:outputGroup = '0' and @amount > 0]/@intermediateExchangeId, '.spold')",
            "name": "//*:activityName/text()",
            "location": "//*:geography/*:shortname/text()",
            "reference product": "//*:intermediateExchange[*:outputGroup = '0' and @amount > 0]/*:name/text()",
            "unit": "//*:intermediateExchange[*:outputGroup = '0' and @amount > 0]/*:unitName/text()",
        },
    }
    ECOSPOLD1_BIO = {
        "expression language": "XPath",
        "labels": {
            "name": "//*:exchange[*:outputGroup = '4' or *:outputGroup = '5']/@name",
            "unit": "//*:exchange[*:outputGroup = '4' or *:outputGroup = '5']/@unit",
            "context": [
                "//*:exchange[*:outputGroup = '4' or *:outputGroup = '5']/@category",
                "//*:exchange[*:outputGroup = '4' or *:outputGroup = '5']/@subCategory",
            ],
        },
    }
    ECOSPOLD2_BIO = {
        "expression language": "XPath",
        "labels": {
            "name": "//*:elementaryExchange/*:name/text()",
            "unit": "//*:elementaryExchange/*:unitName/text()",
            "uuid": "//*:elementaryExchange/@elementaryExchangeId",
            "formula": "//*:elementaryExchange/@formula",
            "context": [
                "//*:elementaryExchange/*:compartment/*:compartment/text()",
                "//*:elementaryExchange/*:compartment/*:subcompartment/text()",
            ],
        },
    }
    ECOSPOLD2_BIO_FLOWMAPPER = {
        "expression language": "XPath",
        "labels": {
            "name": "//*:elementaryExchange/*:name/text()",
            "CAS number": "//*:elementaryExchange/@casNumber",
            "unit": "//*:elementaryExchange/*:unitName/text()",
            "identifier": "//*:elementaryExchange/@elementaryExchangeId",
            "context": [
                "//*:elementaryExchange/*:compartment/*:compartment/text()",
                "//*:elementaryExchange/*:compartment/*:subcompartment/text()",
            ],
        },
    }



from typing import List

from randonneur.utils import apply_mapping


def test_apply_mapping_basic():
    given = {"replace": {}, "disaggregate": {}, "create": {}}


def apply_mapping(migrations: dict, mapping: dict, verbs: List[str]) -> dict:
    """Apply the label changes in `mapping` to the transformations in `migrations`."""
    if "source" in mapping:
        for verb in verbs:
            for transformation in migrations.get(verb, []):
                for key, value in mapping["source"].items():
                    if key in transformation["source"]:
                        transformation["source"][value] = transformation["source"].pop(key)
    if "target" in mapping:
        for verb in verbs:
            if verb == "disaggregate":
                for transformation_list in migrations.get(verb, []):
                    for transformation in transformation_list["targets"]:
                        for key, value in mapping["target"].items():
                            if key in transformation:
                                transformation[value] = transformation.pop(key)
            elif verb in ("create", "delete"):
                continue
            else:
                for transformation in migrations.get(verb, []):
                    for key, value in mapping["target"].items():
                        if key in transformation["target"]:
                            transformation["target"][value] = transformation["target"].pop(key)
    return migrations


{
    "name": "generic-brightway-units-normalization",
    "description": "Standard units normalization used in most Brightway projects",
    "contributors": [
        {"title": "Chris Mutel", "path": "https://chris.mutel.org/", "role": "author"}
    ],
    "created": "2024-07-25T06:47:10.575370+00:00",
    "version": "1.0.0",
    "licenses": [
        {
            "name": "CC BY 4.0",
            "path": "https://creativecommons.org/licenses/by/4.0/",
            "title": "Creative Commons Attribution 4.0 International",
        }
    ],
    "graph_context": ["nodes", "edges"],
    "mapping": {
        "source": {"expression language": "JSONPath", "labels": {"unit": "Node.unit"}},
        "target": {"expression language": "JSONPath", "labels": {"unit": "Node.unit"}},
    },
    "source_id": "bw_interfaces_schemas-1",
    "target_id": "bw_interfaces_schemas-1",
    "homepage": "https://github.com/brightway-lca/bw_interface_schemas",
    "replace": [
        {"source": {"unit": "a"}, "target": {"unit": "year"}},
        {"source": {"unit": "h"}, "target": {"unit": "hour"}},
        {"source": {"unit": "ha"}, "target": {"unit": "hectare"}},
        {"source": {"unit": "hr"}, "target": {"unit": "hour"}},
        {"source": {"unit": "kg"}, "target": {"unit": "kilogram"}},
    ],
}



__all__ = (
    "__version__",
    "migrate_edges",
    "Datapackage",
    "MappingConstants",
    "migrate_edges_with_stored_data",
    "MigrationConfig",
)

__version__ = "0.2.2"

from .constants import MappingConstants
from .datapackage import Datapackage
from .config import MigrationConfig

# from .datasets import migrate_datasets
from .edges import migrate_edges, migrate_edges_with_stored_data



import datetime
from randonneur.validation import DatapackageMetadata, Contributors, MappingFields
from randonneur import Datapackage
import pytest
from pydantic import ValidationError


def test_contributors():
    Contributors(**{"title": "John", "role": "wrangler", "path": "example.com"})
    Contributors(
        **{"title": "John", "role": "wrangler", "path": "example.com", "extra": "ignore"}
    )

    with pytest.raises(ValidationError):
        Contributors(**{"title": "John", "role": "wrangler"})


def test_mapping_fields():
    MappingFields(
        **{"expression language": "foo", "labels": {"a": "something", "b": ["x", "y"]}}
    )

    with pytest.raises(ValidationError):
        MappingFields(
            **{"expression language": "foo", "labels": {"a": "something", "b": 42}}
        )
    with pytest.raises(ValidationError):
        MappingFields(
            **{"expression language": "foo"}
        )


def test_datapackage_metadata_complete():
    DatapackageMetadata(
        name="Foo",
        description="Bar",
        source_id="Something",
        target_id="Something else",
        homepage="https://example.com",
        created=datetime.datetime.now(),
        version="1.0",
        licenses=[{"foo": "bar"}],
        graph_context=['create', 'delete']
    )


def test_datapackage_metadata_partial():
    DatapackageMetadata(
        name="Foo",
        description="Bar",
        version="1.0",
        licenses=[{"foo": "bar"}],
        graph_context=['create', 'delete']
    )


def test_datapackage_metadata_error():
    with pytest.raises(ValidationError):
        DatapackageMetadata(
            name=42,
            description="Bar",
            source_id="Something",
            target_id="Something else",
            homepage="https://example.com",
            created=datetime.datetime.now(),
            version="1.0",
            licenses=[{"foo": "bar"}],
            graph_context=['create', 'delete']
        )
    with pytest.raises(ValidationError):
        DatapackageMetadata(
            name="Foo",
            description="Bar",
            source_id="Something",
            target_id="Something else",
            homepage="https://example.com",
            created=datetime.datetime.now(),
            version="1.0",
            licenses="foo",
            graph_context=['create', 'delete']
        )
    with pytest.raises(ValidationError):
        DatapackageMetadata(
            description="Bar",
            source_id="Something",
            target_id="Something else",
            homepage="https://example.com",
            created=datetime.datetime.now(),
            version="1.0",
            licenses=[{"foo": "bar"}],
            graph_context=['create', 'delete']
        )


def test_validation_integration():
    Datapackage(
        name="Foo",
        description="Bar",
        contributors=[{"title": "John", "role": "wrangler", "path": "example.com"}],
        mapping_source={"expression language": "foo", "labels": {"a": "something", "b": ["x", "y"]}},
        mapping_target={"expression language": "foo", "labels": {"a": "something", "b": ["x", "y"]}},
        source_id="Something",
        target_id="Something else",
        homepage="https://example.com",
        created=datetime.datetime.now(),
        version="1.0",
        licenses=[{"foo": "bar"}],
        graph_context=['create', 'delete']
    )
    Datapackage(
        name="Foo",
        description="Bar",
        contributors=[{"title": "John", "role": "wrangler", "path": "example.com"}],
        mapping_source={"expression language": "foo", "labels": {"a": "something", "b": ["x", "y"]}},
        mapping_target={"expression language": "foo", "labels": {"a": "something", "b": ["x", "y"]}},
    )
    with pytest.raises(ValidationError):
        Datapackage(
            name=42,
            description="Bar",
            contributors=[{"title": "John", "role": "wrangler", "path": "example.com"}],
            mapping_source={"expression language": "foo", "labels": {"a": "something", "b": ["x", "y"]}},
            mapping_target={"expression language": "foo", "labels": {"a": "something", "b": ["x", "y"]}},
            source_id="Something",
            target_id="Something else",
            homepage="https://example.com",
            created=datetime.datetime.now(),
            version="1.0",
            licenses=[{"foo": "bar"}],
            graph_context=['create', 'delete']
        )
    with pytest.raises(ValidationError):
        Datapackage(
            name="Foo",
            description="Bar",
            contributors=[{"title": "John", "role": "wrangler", "path": "example.com"}],
            mapping_source={"labels": {"a": "something", "b": ["x", "y"]}},
            mapping_target={"expression language": "foo", "labels": {"a": "something", "b": ["x", "y"]}},
            source_id="Something",
            target_id="Something else",
            homepage="https://example.com",
            created=datetime.datetime.now(),
            version="1.0",
            licenses=[{"foo": "bar"}],
            graph_context=['create', 'delete']
        )
    with pytest.raises(ValidationError):
        Datapackage(
            name="Foo",
            description="Bar",
            contributors=[{"title": "John", "role": "wrangler"}],
            mapping_source={"expression language": "foo", "labels": {"a": "something", "b": ["x", "y"]}},
            mapping_target={"expression language": "foo", "labels": {"a": "something", "b": ["x", "y"]}},
            source_id="Something",
            target_id="Something else",
            homepage="https://example.com",
            created=datetime.datetime.now(),
            version="1.0",
            licenses=[{"foo": "bar"}],
            graph_context=['create', 'delete']
        )
